"""Module for geojson streaming logic

Makes use of the excelent ijson library to stream and parse into python objects a JSON
document starting at the `features` item. This assumes that a geojson is in the form

::

    {
        "type": "FeatureCollection",
        "features": [
            { ... },
            ...
        ],
        "properties
    }

"""
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union

import geojson
import ijson


class GeoJSONBatchStreamer:
    """Wrapper class around ijson iterable, allowing iteration in batches

    Attributes:
        geojson (Union[str, Path]): Filepath for a valid geojson document.
    """

    def __init__(self, geojson: Union[str, Path]) -> None:
        """
        Constructor for GeoJSONBatchStreamer
        
        Args:
            geojson (Union[str, Path]): Filepath for a valid geojson document. Will
                attempt to convert to a Path object regardless of input type.
        
        Raises:
            FileNotFoundError: If `geojson` does not exist.
        """
        self.geojson = Path(geojson)
        if not self.geojson.exists():
            raise FileNotFoundError(f"file {self.geojson.name} does not exist")

    def stream(
        self, batch: Optional[int] = None, prefix: Optional[str] = None
    ) -> Iterator[geojson.feature.FeatureCollection]:
        """
        Generator method to yield batches of geojson Features in a Feature Collection.
        
        Args:
            batch (Optional[int], optional): The number of features in a single batch. Defaults to 100.
            prefix (Optional[str], optional): The prefix of the element of interest in the
                geojson document. Usually this should be `'features.item'`. Only change this if
                you now what you are doing. See https://github.com/ICRAR/ijson for more info.
                Defaults to `'features.item'`.
        
        Yields:
            (Iterator[geojson.feature.FeatureCollection]): 
                The next batch of features wrapped in a new Feature Collection. This itself is
                just a subclass of a Dict instance, containing typical geojson attributes
                including a JSON array of Features. When `StopIteration` is raised, will yield
                whatever has been gathered so far in the `data` variable to ensure all
                features are collected.
        """
        if batch is None:
            batch = 100
        if prefix is None:
            prefix = "features.item"

        with self.geojson.open("r") as fp:
            features: Iterator[Dict[str, Any]] = ijson.items(fp, prefix)
            try:
                while True:
                    data: List[Dict[str, Any]] = []
                    for _ in range(batch):
                        data.append(next(features))
                    yield geojson.FeatureCollection(data)
            except StopIteration:
                if data:
                    yield geojson.FeatureCollection(data)  # yield remainder of data
                return



import argparse
import logging
import sys
from logging.config import dictConfig
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import geojson
import simplejson as json

from . import __version__
from .geojsplit import GeoJSONBatchStreamer


def input_geojson(args: argparse.Namespace) -> None:
    """Entrypoint function to iter through a valid geojson document and save to multiple files"""

    def gen_filename(
        filename: Path,
        file_count: int,
        width: Optional[int] = None,
        parent: Optional[Path] = None,
    ) -> Path:
        """Generate unique filename according to number of iterations thus far."""
        if width is None:
            width = 4
        if parent is None:
            parent = filename.parent
        elif isinstance(parent, str):
            parent = Path(parent)
        suffix: str = filename.suffix
        stem: str = filename.stem

        return parent / (stem + "_x" + pad(file_count, width) + suffix)

    logger: logging.Logger = logging.getLogger(__name__)
    logger.debug(f"starting splitting with geojson {args.geojson}")
    gj: GeoJSONBatchStreamer = GeoJSONBatchStreamer(args.geojson)

    count: int
    features: geojson.feature.FeatureCollection
    for count, features in enumerate(gj.stream(batch=args.geometry_count)):
        try:
            new_filename: Path = gen_filename(
                gj.geojson, count, width=args.suffix_length, parent=args.output
            )
        except TypeError as e:
            logger.error(f"Could not generate a unique suffix.", exc_info=e)
        try:
            if not args.dry_run:
                if not new_filename.parent.exists():
                    logger.debug(f"creating output directory {args.output}")
                    new_filename.parent.mkdir(parents=True, exist_ok=True)
                with new_filename.open("w") as fp:
                    json.dump(features, fp)
            logger.debug(
                f"successfully saved {len(features['features'])} features to {new_filename}"
            )
        except IOError as e:
            logger.error(f"Could not write features to {new_filename}", exc_info=e)

        # account for 0 based index of enumerate that is required for `pad` method.
        if args.limit is not None:
            if count >= args.limit - 1:
                break


def pad(file_count: int, width: int) -> str:
    """
    Generate sortable alphabetic string to append to filenames.

    Arguments:
        file_count (int): The current file count.
        width (int): 
    """
    logger: logging.Logger = logging.getLogger(__name__)
    alphabet: str = "abcdefghijklmopqrstuvwxyz"

    if file_count >= len(alphabet) ** width:
        logger.error(
            f"Suffix of width of {width} is not enough to generate a unique filename. Increase "
        )
        return

    char_array: List[str] = []
    digit: int
    for digit in range(width - 1, -1, -1):
        digit_cap: int = len(alphabet) ** digit
        idx: int = file_count // digit_cap
        file_count -= idx * digit_cap
        char_array.append(alphabet[idx])

    return "".join(char_array)


def setup_logger() -> None:
    dct: Dict[str, Any] = {
        "version": 1,
        "formatters": {
            "formatter": {
                "class": "logging.Formatter",
                "format": "%(name)s:%(levelname)s:%(message)s",
            }
        },
        "handlers": {
            "stream_handler": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "formatter",
            }
        },
        "root": {"level": "ERROR", "handlers": ["stream_handler"]},
    }

    dictConfig(dct)


def limit_type(x):
    x = int(x)
    if x <= 0:
        raise argparse.ArgumentTypeError("LIMIT must be > 1")
    return x


def setup_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="geojsplit", description="Split a geojson file into many geojson files."
    )
    parser.add_argument("geojson", help="filename of geojson file to split")
    parser.add_argument(
        "-l",
        "--geometry-count",
        type=int,
        help="the number of features to be distributed to each file.",
    )
    parser.add_argument(
        "-a",
        "--suffix-length",
        type=int,
        help="number of characters in the suffix length for split geojsons",
    )
    parser.add_argument(
        "-o", "--output", help="output directory to save split geojsons"
    )
    parser.add_argument(
        "-n",
        "--limit",
        type=limit_type,
        help="limit number of split geojson file to at most LIMIT, with GEOMETRY_COUNT number of features.",
    )
    parser.add_argument(
        "-v", "--verbose", help="increase output verbosity", action="store_true"
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        help="see output without actually writing to file",
        action="store_true",
    )
    parser.add_argument(
        "--version",
        help="show %(prog)s version number",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    return parser


def main(args=None) -> None:
    setup_logger()
    logger: logging.Logger = logging.getLogger(__name__)
    parser: argparse.ArgumentParser = setup_parser()
    args: argparse.Namespace = parser.parse_args(args=args)

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    logger.debug(f"called {__name__} with arguments:")
    for arg_name, arg_value in vars(args).items():
        if arg_value is not None:
            logger.debug(f"{arg_name}: {arg_value}")

    input_geojson(args)
    logger.debug(f"finished splitting geojson")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['geojsplit']

package_data = \
{'': ['*']}

install_requires = \
['geojson>=2.5,<3.0', 'ijson>=2.4,<3.0', 'simplejson>=3.16,<4.0']

extras_require = \
{'docs': ['sphinx>=2.2,<3.0', 'sphinx_rtd_theme>=0.4.3,<0.5.0']}

entry_points = \
{'console_scripts': ['geojsplit = geojsplit.cli:main']}

setup_kwargs = {
    'name': 'geojsplit',
    'version': '0.1.2',
    'description': 'A python implementation of the npm package geojsplit. Used to split GeoJSON files into smaller pieces.',
    'long_description': '# geojsplit\nA python implementation of the node package geojsplit: https://github.com/woodb/geojsplit\n\n[![Build Status](https://travis-ci.com/underchemist/geojsplit.svg?branch=master)](https://travis-ci.com/underchemist/geojsplit)\n[![Documentation Status](https://readthedocs.org/projects/geojsplit/badge/?version=latest)](https://geojsplit.readthedocs.io/en/latest/?badge=latest)\n[![Coverage Status](https://coveralls.io/repos/github/underchemist/geojsplit/badge.svg?branch=master)](https://coveralls.io/github/underchemist/geojsplit?branch=master)\n![PyPI](https://img.shields.io/pypi/v/geojsplit)\n![GitHub](https://img.shields.io/github/license/underchemist/geojsplit)\n\n## Installation\n\n### With poetry\n\nFor an [introduction to poetry](https://dev.to/yukinagae/beginner-guide-on-poetry-new-python-dependency-management-tool-4327/).\n\n```\n$ poetry add geojsplit\n```\n\nwill add geojsplit to your current virtual environment and update your poetry.lock file. If you would like to contribute or develop geojsplit\n\n```\n$ git clone https://github.com/underchemist/geojsplit.git\n$ cd geojsplit\n$ poetry install\n```\n\nYou may need some extra configuration to make poetry play nice with conda virtual environments\n\n```\npoetry config settings.virtualenvs.path <path_to_conda_install>/envs  # tell poetry where you virtualenvs are stored\npoetry config settings.virtualenvs.create 0  # tell poetry not to try to create its own virtualenvs.\n```\n\nSee https://github.com/sdispater/poetry/issues/105#issuecomment-498042062 for more info.\n\n```\n$ poetry config settings.virtualenvs.path $CONDA_ENV_PATH\n$ poetry config settings.virtualenvs.create 0\n```\n\n### With pip\nThough geojsplit is developed using [poetry](https://poetry.eustace.io/) (and as such does not have a setup.py), [pep517](https://www.python.org/dev/peps/pep-0517/) implementation in pip means we can install it directly\n\n```\n$ pip install geojsplit\n```\n\n## Usage\n\nAlthough both the library code and the command line tool of geojsplit are relatively simple, there are use cases for both. You may want to use the backend `GeoJSONBatchStreamer` class directly in order to do more sophisticated manipulations with GeoJOSN documents. As a command line tool geojsplit also works well as a preprocessing step for working with large GeoJSON documents i.e. for piping into GDAL’s ogr2ogr tool.\n\n### As a library\n\nOnce installed, geojsplit can be imported in like\n\n```\nfrom geojsplit import geojsplit\n\ngeojson = geojsplit.GeoJSONBatchStreamer("/path/to/some.geojson")\n\nfor feature_collection in geojson.stream():\n    do_something(feature_collection)\n    ...\n```\n\nIf the `/path/to/some.geojson` does not exists, `FileNotFound` will be raised.\n\nYou can control how many features are streamed into a Feature Collection using the batch parameter (Default is 100).\n\n```\n>>> g = geojson.stream(batch=2)  # instatiate generator object\n>>> data = next(g)\n>>> print(data)\n{"features": [{"geometry": {"coordinates": [[[-118.254638, 33.7843], [-118.254637,\n33.784231], [-118.254556, 33.784232], [-118.254559, 33.784339], [-118.254669,\n33.784338], [-118.254668, 33.7843], [-118.254638, 33.7843]]], "type": "Polygon"},\n"properties": {}, "type": "Feature"}, {"geometry": {"coordinates": [[[-118.254414,\n33.784255], [-118.254232, 33.784255], [-118.254232, 33.784355], [-118.254414,\n33.784355], [-118.254414, 33.784255]]], "type": "Polygon"}, "properties": {}, "type":\n"Feature"}], "type": "FeatureCollection"}\n>>> print(len(data["features"]))\n2\n```\n\nIf your GeoJSON document has a different format or you want to iterate over different elements on your document, you can also pass a different value to the `prefix` keyword argument (Default is `\'features.item\'`). This is an argument passed directly down to a `ijson.items` call, for more information see https://github.com/ICRAR/ijson.\n\n### As a command line tool\n\nAfter installing you should have the geojsplit executable in your `PATH`.\n\n```\n$ geojsplit -h\nusage: geojsplit [-h] [-l GEOMETRY_COUNT] [-a SUFFIX_LENGTH] [-o OUTPUT]\n                [-n LIMIT] [-v] [-d] [--version]\n                geojson\n\nSplit a geojson file into many geojson files.\n\npositional arguments:\ngeojson               filename of geojson file to split\n\noptional arguments:\n-h, --help            show this help message and exit\n-l GEOMETRY_COUNT, --geometry-count GEOMETRY_COUNT\n                        the number of features to be distributed to each file.\n-a SUFFIX_LENGTH, --suffix-length SUFFIX_LENGTH\n                        number of characters in the suffix length for split\n                        geojsons\n-o OUTPUT, --output OUTPUT\n                        output directory to save split geojsons\n-n LIMIT, --limit LIMIT\n                        limit number of split geojson file to at most LIMIT,\n                        with GEOMETRY_COUNT number of features.\n-v, --verbose         increase output verbosity\n-d, --dry-run         see output without actually writing to file\n--version             show geojsplit version number\n```\n\nBy default splitted GeoJSON files are saved as `filename_x<SUFFIX_LENGTH characters long>.geojson`. Default SUFFIX_LENGTH is 4, meaning that 456976 unique files can be generated. If you need more use `-a` or `--suffix-length` to increase this value appropriately.\n\nThe `--geometry-count` flag corresponds to the batch keyword argument for `GeoJSONBatchStreamer.stream` method. Note that if GEOMETRY_COUNT does not divide equally into the number of features in the Feature Collection, the last batch of features will be < GEOMETRY_COUNT.\n\nFinally, to only iterate over the the first n elements of a GeoJSON document, use `--limit`.',
    'author': 'Yann-Sebastien Tremblay-Johnston',
    'author_email': 'yanns.tremblay@gmail.com',
    'url': 'https://github.com/underchemist/geojsplit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
